from .simulator import Simulation
from .info_parser import get_positions
from .read_non_blocking import NonBlockingBytesReader

from subprocess import Popen, PIPE
import sys
import os
import time
import datetime

import argparse

from jelka_validator import DataReader

parser = argparse.ArgumentParser(description="Run jelka simulation.")

parser.add_argument("runner", type=str, nargs="?", help="How to run your program.")
parser.add_argument("target", type=str, help="Your program name.")
parser.add_argument(
    "--positions", type=str, help="File with LED positions. (Leave empty for automatic detection or random.)", required=False
)


def main(header_wait: float = 0.5):
    print(
        f"[SIMULATION] You are executing JelkaSim from '{os.getcwd()}' using Python '{sys.executable}'.",
        file=sys.stderr,
        flush=True,
    )

    args = parser.parse_args()

    cmd = []
    if args.runner:
        if not args.target:
            # TODO: UI
            raise ValueError("[SIMULATION] You must provide a target program.")
    if args.target:
        if args.runner:
            cmd = [args.runner, args.target]
        elif args.target.endswith(".py"):
            cmd = [sys.executable, args.target]
        else:
            cmd = [args.target]
    if cmd == []:
        raise ValueError("[SIMULATION] You must provide a target program. Wait for the next update.")

    # Provide default file locations
    filenames = [
        os.path.join(os.getcwd(), "positions.csv"),
        os.path.join(os.path.dirname(args.target), "positions.csv"),
        os.path.join(os.path.dirname(sys.argv[0]), "positions.csv"),
        os.path.join(os.getcwd(), "../../data/positions.csv"),
        os.path.join(os.path.dirname(args.target), "../../data/positions.csv"),
        os.path.join(os.path.dirname(sys.argv[0]), "../../data/positions.csv"),
    ]

    # Allow specifying custom path
    if args.positions:
        filenames = [args.positions]

    # Try to load positions from various files
    positions = get_positions(filenames)

    print("[SIMULATION] Initializing the simulation window.", file=sys.stderr, flush=True)
    sim = Simulation(positions)
    sim.init()

    print(f"[SIMULATION] Running {cmd} at {datetime.datetime.now()}.", file=sys.stderr, flush=True)

    with Popen(cmd, stdout=PIPE) as p:
        breader = NonBlockingBytesReader(p.stdout.read1)  # type: ignore
        dr = DataReader(breader.start())  # type: ignore
        dr.update()

        t_start = time.time()
        while time.time() - t_start < header_wait and dr.header is None:
            dr.update()
            time.sleep(0.01)

        if dr.header is None:
            raise ValueError(f"[SIMULATION] No header found in the first {header_wait} seconds. Is your program running?")

        while sim.running:
            c = next(dr)
            dr.user_print()
            sim.set_colors(dict(zip(range(len(c)), c)))
            sim.frame()
        breader.close()
        sim.quit()

    print(
        f"[SIMULATION] Finished running at {datetime.datetime.now()} (took {time.time() - t_start:.2f} seconds).",
        file=sys.stderr,
        flush=True,
    )



from numpy import array, cos, sin, random


def proj(point, cam):
    dy = point[1][0] - cam[1]
    if dy != 0:
        return ((point[0][0] - cam[0]) / dy, (point[2][0] - cam[2]) / dy, dy)
    else:
        return (0, 0, 1)


def rotation(phi, tau):
    xy = array([[cos(phi), -sin(phi), 0], [sin(phi), cos(phi), 0], [0, 0, 1]])
    yz = array([[1, 0, 0], [0, cos(tau), -sin(tau)], [0, sin(tau), cos(tau)]])
    return yz @ xy


def draw_lucka(pygame, lucka, screen, size, color, scale):
    w, h = pygame.display.get_surface().get_size()
    pygame.draw.circle(screen, color, (w * lucka[0] * scale + w // 2, (-w * lucka[1] * scale + h // 2)), size + 3)


def draw_line(pygame, p1, p2, screen, color, scale):
    w, h = pygame.display.get_surface().get_size()
    pygame.draw.line(
        screen,
        color,
        (w * p1[0] * scale + w // 2, (-w * p1[1] * scale + h // 2)),
        (w * p2[0] * scale + w // 2, (-w * p2[1] * scale + h // 2)),
    )


class Simulation:
    def __init__(self, smreka) -> None:
        self.running = True
        self.phi, self.tau = 0, 0

        self.smreka = smreka.copy()
        self.points = {i: array([[p[0]], [p[1]], [p[2]]]) for i, p in self.smreka.items()}
        self.colors = {i: (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)) for i in self.smreka}

        self.scale = 1
        self.camera = (0, -500, 100)

    def set_colors(self, colors):
        if not self.running:
            raise InterruptedError("Simulation stopped.")
        try:
            for i, c in colors.items():
                assert all(isinstance(c[i], int) for i in range(3))
                assert all(0 <= c[i] <= 255 for i in range(3))
        except AssertionError:
            self.running = False
            raise ValueError(f"Wrong shape for color: ({i}: {c})") from None  # type: ignore
        self.colors = {pk: tuple(color) for pk, color in colors.items()}

    def init(self):
        import pygame

        self.pygame = pygame
        pygame.init()
        self.screen = pygame.display.set_mode((1280, 720), pygame.RESIZABLE)
        self.clock = pygame.time.Clock()
        pygame.mouse.get_rel()

    def frame(self):
        pygame = self.pygame
        self.w, self.h = pygame.display.get_surface().get_size()
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
            elif event.type == pygame.MOUSEWHEEL:
                self.camera = (self.camera[0], self.camera[1] + event.y * 5, self.camera[2])
            elif event.type == pygame.VIDEORESIZE:
                old_surface_saved = self.screen
                self.screen = pygame.display.set_mode((event.w, event.h), pygame.RESIZABLE)
                self.screen.blit(old_surface_saved, (0, 0))
                del old_surface_saved

        mouse_pressed = pygame.mouse.get_pressed(num_buttons=3)
        key_pressed = pygame.key.get_pressed()
        if mouse_pressed[0] and not key_pressed[pygame.K_LCTRL] and not key_pressed[pygame.K_RCTRL]:
            dphi, dtau = pygame.mouse.get_rel()
            self.phi += dphi / 100
            self.tau += dtau / 100 if dphi == 0 or dtau / dphi > 0.69 else 0
        elif mouse_pressed[0] and (key_pressed[pygame.K_LCTRL] or key_pressed[pygame.K_RCTRL]):
            dx, dz = pygame.mouse.get_rel()
            self.camera = (self.camera[0] - dx / 2, self.camera[1], self.camera[2] + dz / 2)
        else:
            pygame.mouse.get_rel()

        self.screen.fill("black")

        r = rotation(self.phi, self.tau)
        prev = None
        projected = {i: proj(r @ p, self.camera) for i, p in self.points.items()}

        for _, p in sorted(projected.items()):
            if p[2] > 0 and prev and prev[2] > 0:
                draw_line(pygame, p, prev, self.screen, (10, 10, 10), self.scale)
            prev = p

        for i, p in projected.items():
            c = self.colors.get(i, (0, 0, 0))
            draw_lucka(pygame, (p[0], p[1]), self.screen, max(20 / p[2], 1), c, self.scale)

        p = proj(r @ array([[0], [0], [0]]), self.camera)
        draw_lucka(pygame, (p[0], p[1]), self.screen, max(20 / p[2], 1), (0, 255, 0), self.scale)
        pygame.display.flip()
        self.clock.tick(60)  # limits FPS to 60

    def quit(self):
        self.running = False
        self.pygame.quit()



import time

print('#{"version": 0, "led_count": 500, "fps": 60}', flush=True)

for i in range(10):
    print("#" + "ff0000" * 500, flush=True)
    time.sleep(1)
    print("#" + "0000ff" * 500, flush=True)
    time.sleep(1)



import sys

import time
from jelka_validator.datawriter import DataWriter

print("Version:", sys.version, file=sys.stderr)

lc = 300
n = 5400
ds = DataWriter(led_count=lc, fps=60)
t = time.time()
for i in range(n):
    ds.write_frame([(i % 256, i % 256, i % 256)] * lc)
    if i % 100 == 0:
        print(f"Printed {i}-th frame at time {time.time() - t}")



import threading
from time import sleep


class NonBlockingBytesReader:
    def __init__(self, f) -> None:
        self.f = f
        self.result = b""
        self.stop = False

    def start(self):
        self.stop = False
        self.worker = threading.Thread(target=self._f)
        self.worker.start()
        return self

    def _f(self):
        while not self.stop:
            try:
                r = self.f()
            except ValueError:
                break
            self.result += r
            sleep(0.005)

    def __call__(self):
        with threading.Lock():
            r = self.result
            self.result = b""
        return r

    def close(self):
        self.stop = True



import random
import os
import sys


def random_tree(n=500, origin=(0, 0, 0), height=200, max_width=120, min_width=60):
    count = 0
    while count < n:
        x = random.uniform(-max_width, max_width)
        y = random.uniform(-max_width, max_width)
        h = random.uniform(0, height)
        max_w = (height - h) / height * max_width
        min_w = max(max_w - max_width + min_width, 0)
        if min_w**2 <= x**2 + y**2 <= max_w**2:
            count += 1
            yield (x + origin[0], y + origin[1], origin[2] + h)


def get_positions(filenames: "list[str]") -> "dict[int, tuple[float, float, float]]":
    for filename in filenames:
        if not os.path.isfile(filename):
            continue

        with open(filename) as file:
            print(f"[SIMULATION] Loading positions from '{filename}'.", file=sys.stderr, flush=True)

            positions = {}

            for line in file.readlines():
                line = line.strip()
                if line == "":
                    continue
                i, x, y, z = line.split(",")
                positions[int(i)] = (float(x), float(y), float(z))

            return positions

    print("[SIMULATION] No valid file found to load positions from. Using random positions.", file=sys.stderr, flush=True)
    return {i: pos for i, pos in enumerate(random_tree())}
